/*
    a tap to disable/enable mheg
    this merely prevents the software from processing the dsmcc stream
    by zapping the subrouting call.
    Note, since you can enable/disable it on the fly; you can play tricks like:
        enable mheg
        tune to a chennel
        wait a bit to allow the cache to fill up
        disable mheg
        viewing the (now static) pages
    The cache will be deleted when you change channels.

    Use the 'Recall' key to bring up a menu (exit cancels, recall propagates to next tap/firmware).

    this also includes an alternative simple patch to fix the remote key mapping
    when applied, the original key codes are supplied in param2
    the original key codes can be used by applying:
        param1 = (param2) ? param2 : param1;
    the adjusted param1 should be passed on to other taps/firmware

    the firmware will now actually use the returned value, rather than using the original key pressed value.
*/
#include <tap.h>

//Every TAP application needs unique ID and TAP allocations that have same ID cant execute at same time.
//ID consists of DWORD( 4byte ) and the MSB bit should be set by 1.
//The following is how to use in source codes.
#define ID_USERTAP 0xbdb00101
TAP_ID( ID_USERTAP );

//Every TAP application has to include Program name, Author name, and Program description as following
//manners in source code.
TAP_PROGRAM_NAME("mheg_off");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("A TAP to disable/enable MHEG");
TAP_ETCINFO(__DATE__);

//----------------------------------------------------------------------------
//local key defines (missing in key.h)
#define RKEY_TvSTB      0x10022       //was RKEY_TvSat
#define RKEY_Opt        0x10010

#define RKEY_List       0x10018       //shared with RKEY_TvRadio

#define RKEY_Right      0x10015       //shared with RKEY_VolDown
#define RKEY_Left       0x1000f       //shared with RKEY_VolUp
#define RKEY_Up         0x10012       //shared with RKEY_ChUp
#define RKEY_Down       0x1001d       //shared with RKEY_ChDown

#define RKEY_SubMove    0x10041
#define RKEY_FileList   0x10042       //or RKEY_PlayList
#define RKEY_Text       0x10043

#define RKEY_Red        0x1003f
#define RKEY_Green      0x10024
#define RKEY_Yellow     0x10025
#define RKEY_Blue       0x10026

#define RKEY_White      0x1003e
#define RKEY_SubSwap    0x10040

//unmap keys to avoid using them by mistake
#undef  RKEY_TvRadio
#undef  RKEY_Exit
#undef  RKEY_VolUp
#undef  RKEY_VolDown
#undef  RKEY_ChUp
#undef  RKEY_ChDown

//unmapped key values
#define exRKEY_TvRadio  0x1000e       //0x10018 (RKEY_List)    remapped to 0x1000e (RKEY_Uhf)
#define exRKEY_Exit     0x10013       //0x10017 (RKEY_VolDown) remapped to 0x10013 (RKEY_AudioTrk)

#define exRKEY_VolUp    0x10020       //0x10015 (RKEY_Right)   remapped to 0x10020 (RKEY_Fav )
#define exRKEY_VolDown  0x10017       //0x1000f (RKEY_Left)    remapped to 0x10017 (RKEY_Exit)
#define exRKEY_ChUp     0x10021       //0x10012 (RKEY_Up)      remapped to 0x10021 (RKEY_Subt)
#define exRKEY_ChDown   0x10037       //0x1001d (RKEY_Down)    remapped to 0x10037 (RKEY_Teletext)

//----------------------------------------------------------------------------
#define FW_KEY_FUNC_1204            0x8012C9E0
#define FW_PROCESS_DSMCC_FUNC_1204  0x8007b690      //the address for the dsmcc patch
#define FW_PROCESS_DSMCC_DATA_1204  0x0c01ea80      //the data for the dsmcc patch

#define FW_KEY_FUNC_1205            0x8012cB64
#define FW_PROCESS_DSMCC_FUNC_1205  0x8007b690      //the address for the dsmcc patch
#define FW_PROCESS_DSMCC_DATA_1205  0x0c01ea80      //the data for the dsmcc patch

#define FW_KEY_FUNC_1209            0x8012d084      //the address for the key mapping patch
#define FW_PROCESS_DSMCC_FUNC_1209  0x8007b6e4      //the address for the dsmcc patch
#define FW_PROCESS_DSMCC_DATA_1209  0x0c01ea95      //the data for the dsmcc patch

//----------------------------------------------------------------------------
#define pSysID                      0xa3fffffa
#define ptrfix(p) (p + _tap_startAddr)

extern int _appl_version;

typedef struct {
    int  id;
    char str[0x20];
}type_item;

typedef struct {
    int  id;
    int  visible;
    TYPE_Window win;
    int  rgn;
    char *fixme_titlep;
    type_item item[0x10];
}type_menu;

//----------------------------------------------------------------------------
dword fw_key_func;
dword fw_process_dsmcc_func;
dword fw_process_dsmcc_data;

type_menu menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {3, "xxx",          },
        {0, "Disable MHEG", },
        {1, "Enable MHEG",  },
        {2, "Exit Tap",     },
    },
};


int TAP_Main(void);
dword TAP_EventHandler(word event, dword param1, dword param2);
int menu_handler(type_menu *menu, dword param1);
int key_handler(dword param1);
void ShowMenu(type_menu *menu);
void DeleteMenu(type_menu *menu);

int patch_key();
int unpatch_key();
int patch_dsmcc();
int unpatch_dsmcc();

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;
    word SysID = *(word*)pSysID;

    TAP_Print("\nopening %s",__tap_program_name__);

    switch(SysID) {
        case 456:
            break;
        default:
            TAP_Print("\nUnsupported system id - %d", SysID);
            TAP_Print("\nTAP only supports system id 456");
            TAP_Print("\n");
            return 0;
    }
    switch(_appl_version) {
        case 0x1204:
            fw_key_func           = FW_KEY_FUNC_1204;
            fw_process_dsmcc_func = FW_PROCESS_DSMCC_FUNC_1204;
            fw_process_dsmcc_data = FW_PROCESS_DSMCC_DATA_1204;
            break;
        case 0x1205:
            fw_key_func           = FW_KEY_FUNC_1205;
            fw_process_dsmcc_func = FW_PROCESS_DSMCC_FUNC_1205;
            fw_process_dsmcc_data = FW_PROCESS_DSMCC_DATA_1205;
            break;
        case 0x1209:
            fw_key_func           = FW_KEY_FUNC_1209;
            fw_process_dsmcc_func = FW_PROCESS_DSMCC_FUNC_1209;
            fw_process_dsmcc_data = FW_PROCESS_DSMCC_DATA_1209;
            break;
        default:
            TAP_Print("\nUnsupported firmware version - %x", _appl_version);
            TAP_Print("\nTAP only supports firmware versions 5.12.04,5.12.05,5.12.09");
            TAP_Print("\n");
            return 0;
    }

    //apply the key patch
    patch_key();

    menu.visible = 0;
    return 1;                 //0=application mode, 1=tsr mode
}

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    int ret;
    dword state, subState;

    switch (event) {
        case EVT_IDLE:
            break;

        case EVT_KEY:
//TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);

            //fix the key code
            param1 = (param2) ? param2 : param1;

            TAP_GetState(&state, &subState);
            if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                if (menu.visible) { // handling keys when options window is shown
                    param1 = menu_handler(&menu, param1);
                }
                else {
                    param1 = key_handler(param1);
                }
            }
	        break;
    }
    return param1;
}

int menu_handler(type_menu *menu, dword param1)
{
    int selection;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:
				    patch_dsmcc();
				    return 0;
				case 1:
				    unpatch_dsmcc();
				    return 0;
				case 2:
				    unpatch_dsmcc();
				    unpatch_key();          //not much point in removing this
                    TAP_Print("\nclosing %s",__tap_program_name__);
				    TAP_Exit();
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1; // cascades key to other TAP's
        case exRKEY_Exit:
			DeleteMenu(menu);
            return 0; // exit without Cascading
        default:
	        TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
    		return 0;
    }
    return param1;
}


int key_handler(dword param1)
{
    switch(param1) {
        case RKEY_Recall:                 //menu
            ShowMenu(&menu);
            return 0;

        case exRKEY_TvRadio:              //no longer shared with RKEY_List
            break;
        case RKEY_List:                   //now available
            break;

        case exRKEY_Exit:                 //this uses a different keycode
            break;
        case exRKEY_VolUp:                //no longer shared with RKEY_Left
            break;
        case exRKEY_VolDown:              //no longer shared with RKEY_Right
            break;
        case exRKEY_ChUp:                 //no longer shared with RKEY_Up
            break;
        case exRKEY_ChDown:               //no longer shared with RKEY_Down
            break;

        case RKEY_Up:                     //now available
            break;
        case RKEY_Down:                   //now available
            break;
        case RKEY_Left:                   //now available
            break;
        case RKEY_Right:                  //now available
            break;
    }
    return param1;
}

//----------------------------------------------------------------------------
void ShowMenu(type_menu *menu)
{
    int i;

	if (!menu->visible) {
	    menu->rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
	    TAP_Win_SetDefaultColor(&menu->win);
	    TAP_Win_Create(&menu->win, menu->rgn, 250, 210, 200, 80, FALSE, FALSE);
        TAP_Win_SetTitle(&menu->win, ptrfix(menu->fixme_titlep), 0, FNT_Size_1622);
	    for (i=1; i<=menu->item[0].id; i++) {
    	    TAP_Win_AddItem(&menu->win, menu->item[i].str);
	    }
	    TAP_Win_SetSelection(&menu->win, 0);
	    menu->visible = 1;
	}
}

void DeleteMenu(type_menu *menu)
{
	if (menu->visible) {
	    TAP_Win_Delete(&menu->win);
    	TAP_Osd_Delete(menu->rgn);
	    menu->visible = 0;
    }
}

//----------------------------------------------------------------------------
//removing pointless remote key mapping
//8012d084 00 00 30 25   move  $a2, $0                  # change to: 02 c0 30 25   move  $a2, $s6
//8012d088 af c2 00 00   sw    $v0, 0x00($fp)           # change to: 00 40 b0 25   move  $s6, $v0
int patch_key()
{
    if (*(dword *)(fw_key_func+0) != 0x00003025) return -1;
    if (*(dword *)(fw_key_func+4) != 0xafc20000) return -1;

    TAP_Print("\nApplying Key unmapping patch");
    *(dword *)(fw_key_func+0) = 0x02c03025;      //move  $a2, $s6
    *(dword *)(fw_key_func+4) = 0x0040b025;      //move  $s6, $v0
    return 0;
}

int unpatch_key()
{
    if (*(dword *)(fw_key_func+0) != 0x02c03025) return -1;
    if (*(dword *)(fw_key_func+4) != 0x0040b025) return -1;

    TAP_Print("\nRemoving Key unmapping patch");
    *(dword *)(fw_key_func+0) = 0x00003025;      //move  $a2, $0
    *(dword *)(fw_key_func+4) = 0xafc20000;      //sw    $v0, 0x00($fp)
    return 0;
}

//----------------------------------------------------------------------------
//patch out dsmcc handler
//8007b6e4 0c 01 ea 95   jal   0x8007aa54               # process_dsmcc
int patch_dsmcc()
{
    if (*(dword *)(fw_process_dsmcc_func+0) != fw_process_dsmcc_data) return -1;

    TAP_Print("\nApplying MHEG disable patch");
    *(dword *)(fw_process_dsmcc_func+0) = 0x00000000;               //nop
    return 0;
}

int unpatch_dsmcc()
{
    if (*(dword *)(fw_process_dsmcc_func+0) != 0x00000000) return -1;

    TAP_Print("\nRemoving MHEG disable patch");
    *(dword *)(fw_process_dsmcc_func+0) = fw_process_dsmcc_data;    //jal     process_dsmcc
    return 0;
}
//----------------------------------------------------------------------------
